/* drivers/ssi/ads7843.c
 *
 * Driver for the ADS7843 in the Psion 5mx
 * Copyright 2004 Petr Kristan <petr.kristan@epos.cz>
 * Special thanks to Lumir Bezstarosti <lumirb@email.cz>
 *
 * Inspired on adc7843.c, thanks to:
 * Shane R. Nay <shane@minirl.com>
 * Thomas A. de Ruiter <thomas@de-ruiter.cx>
 * Simon Kagstrom <ska@bth.se>
 *
 */

/* /dev/tpanel 10 11
 * Every read from tpanel driver should produce *3* shorts, first is
 * pen status, either 1 or 0.  The other in order are x coord, and
 * then y coord.  If less than 3 shorts are requested, then it will
 * give the requested amount of information starting at the top, and
 * will abandon the rest, and move it's pointer up to the next set
 * of coordinates.
 * It will only return one pen up info set.
 *
 * /proc/apm
 * Returns AC - online/offline, and battery capacity.
 * Highly depends on bat_* parameters
 *
 * /proc/battery
 * Returns main/backup battery voltage
 * accepts battery linearization parameters:
 *   bat_capacity bat_avg_current bat_min_voltage bat_max_voltage
 *
 * EXAMPLE
 * echo 2300 100 2450 2600 > /proc/battery
 *
 * Means: batt capacity 2300mAh, 100mA avg current, 2450mV min batt voltage,
 * 2600mV max batt voltage 
 * these params are used for stamina determination in /proc/apm
 *
 */

/* TODO
 *
 * 1. backup battery voltage does not work for me and i do not know why :(
 * 2. maybe some pen filtration
 * 3. apmd does not work, need to make fake /dev/apm_bios
 * 4. move /proc/battery do /proc/psionw, but modularity must be preserved
 * 5. fix gpm
 *
 */


#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/miscdevice.h>
#include <linux/sched.h>
#include <linux/tqueue.h>
#include <linux/proc_fs.h>
#include <linux/poll.h>
#include <linux/interrupt.h>
#include <linux/timer.h>

#include <asm/uaccess.h>
#include <asm/hardware/psionw.h>
#include <asm/mach-types.h>
#include <asm/io.h>
#include <asm/irq.h>

#include "ssi_dev.h"
#include "ssi_bus.h"

extern struct ssi_bus psionw_ssi5mx_bus;
static struct ssi_dev ads7843_dev;

#define TPANEL_BUF_SIZE 32 //Must be power of 2

static wait_queue_head_t wait_q;
static struct tasklet_struct ads_rcv_t, tpanel_irq_t;

static int pen_irq_disabled;
static int pen_down;
static struct timer_list sample_timer, settle_timer;

//request bit queue
static unsigned long req_q;
static int req_q_pos;

static int tpanel_readers;
//tpanel circular buffer
static u_int tpanel_buf[TPANEL_BUF_SIZE][2]; //x, y
static int tpanel_buf_head, tpanel_buf_size;

//battery
static u_int battery[2];
static int battery_ready;

static struct fasync_struct *fasync;

//rcv
static u_int rcv_buf[2];
static int rcv_req_type;


//Assume default NiMh battery
static int bat_capacity = 1300; //[mAh]
static int bat_avg_current = 100; //[mA]
static int bat_min_voltage = 2400; //[mV]
static int bat_max_voltage = 2600; //[mV]

static struct proc_dir_entry *battery_proc_entry;

#define BAT_OFF 1800 //1.8V psion is switched off

#define VERSION "1.0"
#define REQ_BATT 0
#define REQ_TPANEL 1
#define BAT_VREF 4500 //mV
#define SAMPLE_INTERVAL HZ/10
#define SETTLE_TIME HZ/50
#define FILTER_PEN_SPEED_Y 4000/SAMPLE_INTERVAL
#define FILTER_PEN_SPEED_X 2000/SAMPLE_INTERVAL

static int ads7843_send_req(int req_type)
{
	u_int flags;


	if (req_q_pos>=sizeof(req_q)*8)
		printk("ads7843: req_q overflow\n");

	//ssi_select_device(&psionw_ssi5mx_bus, &ads7843_dev);
	save_flags(flags);
	cli();
	switch (req_type) {
	case REQ_TPANEL:
		ssi_transmit_data(&ads7843_dev, 0xd000); //X
		ssi_transmit_data(&ads7843_dev, 0x9000); //Y
		break;
	case REQ_BATT:
		ssi_transmit_data(&ads7843_dev, 0xa400); //Main battery
		/*
		 * FIXME This code 0xe4 for backup battery does not work for me.
		 * I do not know why?
		 * Return is always zero. In epoc, backup battery voltage works good.
		 */
		ssi_transmit_data(&ads7843_dev, 0xe400); //Backup battery
		break;
	default:
		printk("ads7843: invalid send req_type\n");
		break;
	}
	//enqueue request
	req_q |= req_type << req_q_pos++;
	restore_flags(flags);

	return 0;
}

static inline void enqueue_tpanel_buf(u_int *b)
{
	//enqueue tpanel result
	tpanel_buf[tpanel_buf_head][0] = b[0];
	tpanel_buf[tpanel_buf_head][1] = b[1];
	tpanel_buf_head++;
	tpanel_buf_head &= (TPANEL_BUF_SIZE-1);
	if (tpanel_buf_size < TPANEL_BUF_SIZE-1)
		tpanel_buf_size++;
}

static int ads7843_store_result(int req_type, u_int *b)
{
	u_int flags;

	save_flags(flags);
	cli();

	switch (req_type) {
	case REQ_TPANEL:
		enqueue_tpanel_buf(b);
		break;
	case REQ_BATT:
		//store battery result
		battery[0]=b[0];
		battery[1]=b[1];
		battery_ready++;
		break;
	default:
		printk("ads7843: invalid store req_type\n");
		break;
	}

	restore_flags(flags);

	return 0;
}

static inline void dequeue_tpanel_buf(u_int *b)
{
	int i;

	i = tpanel_buf_head - tpanel_buf_size;
	if (i<0)
		i = TPANEL_BUF_SIZE + i;
	if (tpanel_buf_size)
		tpanel_buf_size--;
	b[0] = tpanel_buf[i][0];
	b[1] = tpanel_buf[i][1];
}

static int ads7843_retrieve_result(int req_type, u_int *b)
{
	u_int flags;

	save_flags(flags);
	cli();

	switch (req_type) {
	case REQ_TPANEL:
		//dequeue tpanel result
		if (!tpanel_buf_size) { //Buffer is empty
			restore_flags(flags);
			return 1;
		}
		dequeue_tpanel_buf(b);
		break;
	case REQ_BATT:
		//retrieve battery result
		b[0]=battery[0];
		b[1]=battery[1];
		battery_ready=0;
		break;
	default:
		printk("ads7843: invalid retrieve req_type\n");
		break;
	}

	restore_flags(flags);

	return 0;
}

/* Recieve byte from SSI bus, ISR no concurrency*/
static void ads7843_dev_rcv(struct ssi_dev* dev, u_int dat)
{
	static u_int buf[2];
	static int cnt;

	//printk("ads7843: rcv: %x\n", dat);

	buf[cnt++] = dat;
	if (cnt==2) {
		cnt=0;

		//dequeue request
		rcv_req_type = req_q & 0x1;
		if (!req_q_pos)
			printk("ads7843: req_q underflow\n");
		req_q_pos--;
		req_q >>= 1;

		//Only 12 bits and second word is inverted, I do not know why
		buf[0] &= 0xfff;
		if (rcv_req_type==REQ_TPANEL) {
			buf[1] = ~ buf[1] & 0xfff;
			pen_down = ( (!buf[0] || !buf[1]) ? 0 : 1 );
		}
		else
			buf[1] = buf[1] & 0xfff;

		rcv_buf[0] = buf[0];
		rcv_buf[1] = buf[1];

		tasklet_schedule(&ads_rcv_t);
	}
}

static void ads_rcv_bh(unsigned long data)
{
	ads7843_store_result(rcv_req_type, rcv_buf);

	kill_fasync(&fasync, SIGIO, POLL_IN);
	wake_up(&wait_q);

	//delay enable tpanel irq
	mod_timer(&sample_timer, jiffies + SAMPLE_INTERVAL);
}

static int ads7843_read_battery(u_int *b, u_int *ac)
{
	int retval;

	ads7843_send_req(REQ_BATT);
	retval = wait_event_interruptible(wait_q, battery_ready);
	if (retval)
		return retval;
	ads7843_retrieve_result(REQ_BATT, b);

	*ac = (psionw_readb(PWRSR) & DCDET) ? 1 : 0;

	return 0;
}

static int tpanel_open(struct inode *node, struct file *file)
{
	tpanel_readers++;
	return 0;
}

static int tpanel_read(struct file *file, char *buf, size_t size, loff_t *pos)
{
	char ret[6];
	u_int b[2];
	int retval;

	int cnt=size<sizeof(ret) ? size : sizeof(ret);

	if (! tpanel_buf_size && (file->f_flags & O_NONBLOCK))
		return -EAGAIN;

	retval = wait_event_interruptible(wait_q, tpanel_buf_size);
	if (retval)
		return retval;
	ads7843_retrieve_result(REQ_TPANEL, b);

	ret[0]=0;
	if (!b[0] || !b[1]) {
		ret[1]=0;

		ret[2]=0;
		ret[3]=0;
		ret[4]=0;
		ret[5]=0;
	} else
		ret[1]=1;
	ret[2]=b[0]>>8;
	ret[3]=b[0] & 0xff;
	ret[4]=b[1]>>8;
	ret[5]=b[1] & 0xff;
	if (copy_to_user(buf, &ret[0], cnt))
		return -EFAULT;
	return cnt;
}

static unsigned int tpanel_poll(struct file *file, poll_table *wait)
{
	poll_wait(file, &wait_q, wait);
	if (tpanel_buf_size)
		return POLLIN | POLLRDNORM;

	return 0;
}

static int tpanel_fasync(int fd, struct file *file, int on)
{
	int retval;
	retval = fasync_helper(fd, file, on, &fasync);
	if (retval)
		return retval;

	return 0;
}

static int tpanel_release(struct inode *inode, struct file *file)
{
	tpanel_fasync(-1, file, 0);
	tpanel_readers--;

	return 0;
}

static struct file_operations ads7843_fops = {
	read:     tpanel_read,
	open:     tpanel_open,
	release:  tpanel_release,
	poll:     tpanel_poll,
	fasync:   tpanel_fasync,
	};

static struct miscdevice ads7843_tpanel = {
		PS5MX_TPANEL_MINOR, "ads7843", &ads7843_fops
	};

static int apm_proc_read(char *buf, char **start, off_t offs, int count, int *eof, void *data)
{
	int len;
	u_int b[2], ac;
	int mv, cap, min;

	ads7843_read_battery(b, &ac);
	mv = BAT_VREF*b[0]/4096;

	cap = (mv - bat_min_voltage)*100/(bat_max_voltage - bat_min_voltage);
	if (cap<0)
		cap=0;
	if (cap>100)
		cap=100;

	min = bat_capacity/bat_avg_current*cap/100*60;

	len = snprintf(buf, count, "%s 1.0 0x3 0x%02x 0x00 0x01 %d%% %d min\n",
		       VERSION, ac, cap, min);

	*eof = 1;
	return len;
}

static char *batt_rem(int mv)
{
	static char rem[][8] = {"Good", "Low", "Replace", "Off"};
	if (mv>2200)
		return rem[0];
	if (mv>2100)
		return rem[1];
	if (mv>BAT_OFF)
		return rem[2];
	return rem[3];
}

static int batt_proc_read(char *buf, char **start, off_t offs, int count, int *eof, void *data)
{
	int len;
	int b[2], ac;
	int mv[2];
	char *rem[2];

	ads7843_read_battery(b, &ac);

	mv[0] = BAT_VREF*b[0]/4096;
	mv[1] = BAT_VREF*b[1]/4096;
	rem[0] = batt_rem(mv[0]);
	rem[1] = batt_rem(mv[1]);

	len = snprintf(buf, count, "AC: %d\nMain: (%d) %d.%dV %s\nBackup: (%d) %d.%dV %s\n\nCapacity: %dmAh\nAvg current: %dmA\nLin min voltage: %dmV\nLin max voltge: %dmV\n",
		       ac,
		       b[0], mv[0]/1000, mv[0]%1000, rem[0],
		       b[1], mv[1]/1000, mv[1]%1000, rem[1],
           bat_capacity,
           bat_avg_current,
           bat_min_voltage,
           bat_max_voltage
        );

	*eof = 1;
	return len;
}

static int str_to_int(const char *buf, int start, int end)
{
  int i;
  int dec = 1;
  int ret = 0;

  for (i = end; i >= start; i--) {
    if ( (buf[i] >= '0') && (buf[i] <= '9') ) {
      ret += dec*(buf[i]-'0');
      dec *= 10;
    }
  }
  return ret;
}

//format: bat_capacity bat_avg_current bat_min_voltage bat_max_voltage
#define APM_WRITE_BUF_SIZE 32

static int batt_proc_write(struct file *file, const char *buf, unsigned long count, void *data)
{
  unsigned char buffer[APM_WRITE_BUF_SIZE];
  int i;
  int is = 0, cnt = 0;
  int val;

  if ( (count < 1) || (count >= APM_WRITE_BUF_SIZE))
    return -EINVAL;

  if (copy_from_user(&buffer, buf, count))
    return -EFAULT;

  for (i = 0; i <= count; i++ ) {
    if ( (i == count) || (buffer[i] == ' ') ) {
      val = str_to_int(buffer, is, i-1);
      switch (cnt++) {
        case 0: 
          bat_capacity = val;
          break;
        case 1:
          bat_avg_current = val;
          break;
        case 2:
          bat_min_voltage = val;
          break;
        case 3:
          bat_max_voltage = val;
          break;
      }
      is = i+1;
    }
  }
  return count;
}

/* Initialization routine called by bus when registered device */
static int ads7843_dev_init(struct ssi_dev* dev)
{
	return 0;
}

static void pen_down_int(int irq, void *dev_id, struct pt_regs *regs)
{
	if (!tpanel_readers)
		return;

	//This interrupt is too fast - now disabled and enabled in adc int. handler
	pen_irq_disabled++;
	disable_irq_nosync(IRQ_EINT3);

	if (! pen_down) {
		pen_down++;
		mod_timer(&settle_timer, jiffies + SETTLE_TIME);
	}
}

static void tpanel_sample(unsigned long data)
{
	ads7843_send_req(REQ_TPANEL);
}

static void settle_timer_func(unsigned long data)
{
	tasklet_schedule(&tpanel_irq_t);
}

static void sample_timer_func(unsigned long data)
{
	if (pen_irq_disabled) {
		pen_irq_disabled=0; //Before enable irq!!
		enable_irq(IRQ_EINT3);
	}

	if (pen_down)
		tasklet_schedule(&tpanel_irq_t);
}

static struct ssi_dev ads7843_dev = {
	name:		"ADS7843",
	id:		1,
	proto:		SSI_MICROWIRE,
	cfglen:		8,
	framelen:	12,
	clkpol:		0,
	clkfreq:	7200, /*7.2kHz-1.8432Mhz*/
	init: ads7843_dev_init,
	rcv: ads7843_dev_rcv,
	bus: NULL,
};

static int __init ads7843_init(void)
{
	int retval;

	retval = ssi_register_device(&psionw_ssi5mx_bus, &ads7843_dev);
	if (retval < 0)
		return retval;

	ssi_select_device(&psionw_ssi5mx_bus, &ads7843_dev);

	retval = request_irq(IRQ_EINT3, pen_down_int, SA_SAMPLE_RANDOM,
			     "ads7843", &ads7843_dev);
	if (retval)
		return retval;

	SET_MODULE_OWNER(&ads7843_fops);

	init_waitqueue_head(&wait_q);

	tpanel_irq_t.func=tpanel_sample;
	ads_rcv_t.func=ads_rcv_bh;

	misc_register(&ads7843_tpanel);
	//create_proc_read_entry("battery", 0444, NULL, batt_proc_read, NULL);
  battery_proc_entry = create_proc_entry("battery", 0444, NULL);
  battery_proc_entry->read_proc = &batt_proc_read;
  battery_proc_entry->write_proc = &batt_proc_write;

	create_proc_read_entry("apm", 0444, NULL, apm_proc_read, NULL);

	init_timer(&sample_timer);
	sample_timer.function = sample_timer_func;
	sample_timer.expires = jiffies - 1; //do not exec event now
	add_timer(&sample_timer);

	init_timer(&settle_timer);
	settle_timer.function = settle_timer_func;
	settle_timer.expires = jiffies - 1; //do not exec event now
	add_timer(&settle_timer);

	return 0;
}

static void __exit ads7843_exit(void)
{
	free_irq(IRQ_EINT3, &ads7843_dev);
	del_timer(&sample_timer);
	del_timer(&settle_timer);

	ssi_select_device(&psionw_ssi5mx_bus, NULL);

	misc_deregister(&ads7843_tpanel);
	remove_proc_entry("battery", NULL);
	remove_proc_entry("apm", NULL);

	ssi_unregister_device(&ads7843_dev);
}

module_init(ads7843_init);
module_exit(ads7843_exit);

MODULE_AUTHOR("Petr Kristan <petr.kristan@epos.cz>");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("ADS7843 Psion touch screen and battery driver");
